const ts = require("typescript");

const generalLoosenRules = {
  "unicorn/consistent-function-scoping": "off",
  "unicorn/error-message": "off",
  "sonarjs/no-identical-conditions": "off",
  "sonarjs/no-identical-expressions": "off",
  "sonarjs/no-identical-functions": "off",
};

const loosenTsRules = {
  "@typescript-eslint/no-unsafe-argument": "off",
  "@typescript-eslint/no-unsafe-assignment": "off",
  "@typescript-eslint/no-unsafe-call": "off",
  "@typescript-eslint/no-unsafe-member-access": "off",
  "@typescript-eslint/no-unsafe-return": "off",
  "@typescript-eslint/require-await": "off",
  "@typescript-eslint/await-thenable": "off",
  "@typescript-eslint/no-explicit-any": "off",
  "@typescript-eslint/restrict-plus-operands": "off",
  "@typescript-eslint/restrict-template-expressions": "off",
};

const loosenReactRules = {
  // loosen some rules
  "react/prop-types": "off",
  "react/react-in-jsx-scope": "off",
};

module.exports = function buildConfig(
  env,
  { nextRules, tsRules, reactRules, defaultRules }
) {
  const config = {
    extends: ["eslint:recommended"],
    env: {
      es2022: true,
      browser: env.react,
      node: true,
    },
    // future: add @brettz9, css-modules, eslint-comments plugins
    plugins: ["sonarjs", "unicorn", "import", "promise"],
    settings: {
      "import/resolver": {
        [require.resolve("eslint-import-resolver-node")]: {
          extensions: [".js", ".jsx"],
        },
      },
    },
    rules: defaultRules,
    overrides: [],
  };

  const loosenRules = {
    ...generalLoosenRules,
    ...(env.react ? loosenReactRules : {}),
    ...(env.typescript ? loosenTsRules : {}),
  };

  if (env.react) {
    Object.assign(config, {
      plugins: [...config.plugins, "react", "react-hooks"],
      rules: {
        ...config.rules,
        ...reactRules,
      },
      settings: {
        ...config.settings,
        react: {
          version: "detect",
        },
      },
      overrides: [
        ...config.overrides,
        {
          files: ["*.stories.*", "*.test.*"],
          rules: loosenRules,
        },
      ],
    });
  }

  if (env.next) {
    Object.assign(config, {
      plugins: [...config.plugins, "@next/next"],
      rules: {
        ...config.rules,
        ...nextRules,
      },
    });
  }

  // order is important here
  if (env.jest) {
    Object.assign(config, {
      plugins: [...config.plugins, "jest"],
      overrides: [
        ...config.overrides,
        {
          extends: ["plugin:jest/recommended"],
          files: ["__test__/**", "test/**", "*.spec.*", "*.test.js"],
          env: {
            "jest/globals": true,
          },
          rules: {
            ...loosenRules,
            // sometimes there are problems with version detection causing this rule to crash ESLint
            "jest/no-deprecated-functions": "off",
          },
        },
      ],
    });
  }

  if (env.typescript) {
    let tsconfigPath;

    try {
      tsconfigPath = ts.findConfigFile(
        process.cwd(),
        ts.sys.fileExists,
        "tsconfig.json"
      );
      const tsConfigFile = ts.readConfigFile(tsconfigPath, ts.sys.readFile);
      if (tsConfigFile.config) {
        // this are type-aware checks that require tsconfig presence
        Object.assign(tsRules, {
          "@typescript-eslint/await-thenable": "warn",
          "@typescript-eslint/no-floating-promises": "error",
          "no-implied-eval": "off",
          "@typescript-eslint/no-implied-eval": "error",
          "@typescript-eslint/no-misused-promises": "warn",
          "@typescript-eslint/no-unnecessary-type-assertion": "warn",
          "require-await": "off",
          "@typescript-eslint/require-await": "error",
          "@typescript-eslint/unbound-method": [
            "error",
            { ignoreStatic: true },
          ],
        });
      }
      if (config?.compilerOptions?.strict) {
        Object.assign(tsRules, {
          "@typescript-eslint/restrict-plus-operands": "warn",
          "@typescript-eslint/no-explicit-any": "error",
          "@typescript-eslint/no-unsafe-argument": "error",
          "@typescript-eslint/no-unsafe-assignment": "error",
          "@typescript-eslint/no-unsafe-call": "error",
          "@typescript-eslint/no-unsafe-member-access": "error",
          "@typescript-eslint/no-unsafe-return": "error",
          "@typescript-eslint/restrict-template-expressions": "error",
        });
      }
    } catch {
      void 0;
    }

    Object.assign(config, {
      plugins: [...config.plugins, "@typescript-eslint"],
      parser: "@typescript-eslint/parser",
      parserOptions: {
        sourceType: "module",
        project: tsconfigPath,
        ecmaFeatures: {
          jsx: env.react,
        },
        warnOnUnsupportedTypeScriptVersion: true,
      },
      settings: {
        ...config.settings,
        "import/resolver": {
          ...config.settings["import/resolver"],
          [require.resolve("eslint-import-resolver-typescript")]: {
            alwaysTryTypes: true,
          },
        },
      },
      rules: {
        ...config.rules,
        ...tsRules,
      },
      overrides: [
        ...config.overrides,
        // ESLint can only apply one override per file, and takes first one matched,
        // so multiple overrides should be combined.
        // order is important.
        {
          files: [
            "*.stories.ts",
            "*.stories.tsx",
            "*.stories.mts",
            "*.stories.cts",
          ],
          rules: {
            ...loosenRules,
          },
        },
        {
          files: [
            "*.test.ts",
            "*.test.tsx",
            "*.test.mts",
            "*.test.cts",
            "__test__/**/*.ts",
            "__test__/**/*.tsx",
            "__test__/**/*.mts",
            "__test__/**/*.cts",
          ],
          extends: ["plugin:jest/recommended"],
          env: {
            "jest/globals": true,
          },
          rules: {
            ...loosenRules,
            "jest/no-deprecated-functions": "off",
          },
        },
        {
          files: ["*.ts", "*.tsx", "*.mts", "*.cts"],
          rules: {
            ...tsRules,
          },
        },
      ],
    });
  }

  return config;
};
